@@ -12,6 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import fcntl
 import sys
 import time
 import uuid
@@ -23,6 +24,8 @@ import errno
 from SocketServer import (ThreadingUnixStreamServer,
                           StreamRequestHandler)
 from io import BytesIO
+from StringIO import StringIO
+import traceback
 
 import pysource
 from pysource import config
@@ -69,9 +72,12 @@ class RequestHandler(StreamRequestHandler):
                 res_payload = {}
             else:
                 res_payload = _handle(**body)
-        except Exception, e:
+        except Exception:
             res_status = RESPONSE_STATUS_ERROR
-            res_payload = {'error': str(e)}
+            error = StringIO()
+            error.write('daemon: ')
+            traceback.print_exc(file=error)
+            res_payload = {'error': error.getvalue()}
 
         if self.pipe_control_handler:
             response_file = self.pipe_control_handler.wfile
@@ -204,6 +210,8 @@ class PipeControlledInputSocket(PipeControlledBaseSocket):
         self.bytes_read = 0
         self.total_bytes = None
         self.done = False
+        self.read_sockets = self.sockets
+        self.control_consumed = False
 
     def p(self, message):
         if DEBUG and ((DEBUG_CLIENT and not request_context.piped) or
@@ -216,21 +224,22 @@ class PipeControlledInputSocket(PipeControlledBaseSocket):
         self.p('read length={}, ensure_length={}'.format(
             length, ensure_length))
         if self.done:
-            return None
+            return ''
         result = BytesIO()
         while True:
             self.p('loop')
             if self.bytes_read == self.total_bytes:
                 break
-            readable, _, _ = select.select(self.sockets, [], [], 0.5)
+            timeout = 30 if ensure_length else 0.5
+            readable, _, _ = select.select(self.read_sockets, [], [], timeout)
             if self.control_socket in readable:
-                control_code = int(self.control_file.readline())
-                self.p('control_code={}'.format(control_code))
-                if control_code == CONTROL_SOCKET_ERROR_CODE:
+                self.read_sockets = [self.data_socket]
+                total_bytes = self.consume_control()
+                if total_bytes == -1:  # error
                     self.total_bytes = self.bytes_read
                     break
                 else:
-                    self.total_bytes = int(self.control_file.readline())
+                    self.total_bytes = total_bytes
                 self.p('total_bytes={}'.format(self.total_bytes))
             if self.data_socket in readable:
                 to_read = 1024 if length <= 0 else length - result.tell()
@@ -259,12 +268,33 @@ class PipeControlledInputSocket(PipeControlledBaseSocket):
             self.done = True
         self.p('out done={}, bytes_read={}, total_bytes={}'.format(
             self.done, self.bytes_read, self.total_bytes))
-        return result.getvalue()
+        value = result.getvalue()
+        if len(value) == 0:
+            if self.done:
+                return ''
+            else:
+                return None
+        else:
+            return result.getvalue()
+
+    def consume_control(self):
+        if self.control_consumed:
+            return -2
+        self.control_consumed = True
+        self.control_socket.setblocking(1)
+        try:
+            control_code = int(self.control_file.readline())
+            self.p('control_code={}'.format(control_code))
+            if control_code == CONTROL_SOCKET_ERROR_CODE:
+                return -1
+            else:
+                return int(self.control_file.readline())
+        finally:
+            self.control_socket.setblocking(0)
 
     def close(self):
         self.p('read_close done={}, bytes_read={}, total_bytes={}'.format(
             self.done, self.bytes_read, self.total_bytes))
-        pass
 
 
 class PipeControlledOutputSocket(PipeControlledBaseSocket):
@@ -336,54 +366,38 @@ def do_piped_client_request(req_type, payload):
                 pipe_control_handler.rfile,
                 pipe_control_handler.conn)
 
+            stdin_fd = sys.stdin.fileno()
             stdin_buf = 16 * 1024
+            fl = fcntl.fcntl(stdin_fd, fcntl.F_GETFL)
+            fcntl.fcntl(stdin_fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
+
+            data_sock = sock
             data_sock_buf = 1024
-            stdin = sys.stdin
-            data_sock = sock
             control_soc = pipe_control_handler.conn
-            read_sockets = [stdin, data_sock, control_soc]
             control_soc_written = False
-            stdin_done = False
-            stdin_done_handled = False
+
+            read_sockets = [stdin_fd, data_sock, control_soc]
 
-            count = 0
             while True:
-                if count % 1000 == 0:
-                    p('loop count={}'.format(count))
-                count += 1
-                readable = select.select(read_sockets, [], [], 0.01)[0]
-                if stdin in readable:
-                    buf = stdin.read(stdin_buf)
-                    if not buf:
-                        stdin_done = True
-                    else:
-                        p('stdin readable')
-                        piped_output.write(buf)
-                        p('write_buf ended')
-                else:
-                    stdin_done = True
-                if not stdin_done_handled and stdin_done:
-                    piped_output.close()
-                    p('piped_output_close')
-                    read_sockets = [control_soc, data_sock]
-                    stdin_done_handled = True
+                readable = select.select(read_sockets, [], [], 30)[0]
+                if stdin_fd in readable:
+                    stdin_read = os.read(stdin_fd, stdin_buf)
+                    if stdin_read == '':
+                        piped_output.close()
+                        read_sockets = [data_sock, control_soc]
+                    elif stdin_read is not None:
+                        piped_output.write(stdin_read)
                 if control_soc in readable:
                     control_soc_written = True
-                    p('control_sock_written=True')
-                if data_sock in readable or control_soc in readable:
+                if data_sock in readable:
                     if control_soc_written:
-                        p('piped_input_read()')
                         buf = piped_input.read()
                     else:
-                        p('piped_input_read(data_sock_buf)')
                         buf = piped_input.read(data_sock_buf,
                                                ensure_length=False)
-                    if buf is None:
-                        p('buf is None')
+                    if buf is '':
                         break
-                    p('read ended len={}'.format(len(buf)))
-                    if len(buf) > 0:
-                        p('stdout.write len={}'.format(len(buf)))
+                    elif buf is not None:
                         sys.stdout.write(buf)
                         try:
                             sys.stdout.flush()
@@ -395,12 +409,10 @@ def do_piped_client_request(req_type, payload):
                                                      'terminated.')
                             else:
                                 raise
-                    else:
-                        p('stdout.write NO_WRITE')
                 if control_soc_written:
                     break
 
-            p('piped_input_close')
+            piped_input.consume_control()
             piped_input.close()
             pipe_control_handler.conn.setblocking(1)
             return pipe_control_handler
